# -*- coding: utf-8 -*-
"""
Pydoas high level test module

Author: Jonas Gliß
Email: jonasgliss@gmail.com
License: GPLv3+
"""
import pytest
import pydoas
import numpy.testing as npt

@pytest.fixture(scope="module")
def example_dataset():
    files, path = pydoas.get_data_files("doasis")
    ### Specify the the import details
    
    # here, 3 x SO2 from the first 3 fit scenario result files (f01, f02, f03) 
    # BrO from f04, 2 x O3 (f02, f04) and OClO (f04)
    import_dict = {'so2'  : ['SO2_Hermans_298_air_conv',
                            ['f01','f02','f03']],
                   'bro'  : ['BrO_Wil298_Air_conv',['f04']],
                   'o3'   : ['o3_221K_air_burrows_1999_conv', ['f02', 'f04']],
                   'oclo' : ['OClO_293K_Bogumil_2003_conv', ['f04']]}
     ### Specify the default fit scenarios for each species
    
    # After import, the default fit scenarios for each species are used
    # whenever fit scenarios are not explicitely specified
    default_dict = {"so2"  :   "f03",
                    "bro"  :   "f04",
                    "o3"   :   "f04",
                    "oclo" :   "f04"}
    
    # Device ID of the spectrometer (of secondary importance)
    dev_id = "avantes"
    
    # Data import type (DOASIS result file format)
    res_type = "doasis"
    #: Create import setup object
    stp = pydoas.\
        dataimport.ResultImportSetup(path, 
                                     result_import_dict=import_dict,
                                     default_dict=default_dict, 
                                     meta_import_info=res_type,
                                     dev_id=dev_id)

    #: Create Dataset object for setup...
    ds = pydoas.analysis.DatasetDoasResults(stp)
    ds.load_raw_results()
    return ds

def test_dataset_lowlevel(example_dataset):
    ds = example_dataset
    raw = ds.raw_results
    
    npt.assert_array_equal([sum([x in raw.keys() 
                                for x in ["f01", "f02", "f03", "f04"]]),
                            ds.get_default_fit_id("so2"),
                            ds.get_default_fit_id("bro"),
                            ds.get_default_fit_id("oclo")],
                            [4, "f03", "f04", "f04"])

    so2_fit_ids = ds.import_info["so2"][1]
    vals=[]
    
    for fid in so2_fit_ids:
        vals.append(raw[fid]["so2"].mean())
        vals.append(raw[fid]["so2_err"].mean())
        
    o3_fit_ids = ds.import_info["o3"][1]
    for fid in o3_fit_ids:
        vals.append(raw[fid]["o3"].mean())
        vals.append(raw[fid]["o3_err"].mean())
        
    vals.extend([raw["f01"]["delta"].mean(),
                 raw["f01"]["chi2"].mean(),
                 raw["f01"]["texp"].sum(),
                 raw["f01"]["lon"].mean(),
                 raw["f01"]["lat"].mean(),
                 raw["f01"]["rms"].mean(),])

    npt.assert_allclose(actual=vals,
                        desired=[1.0835821818181818e+18, 
                                 1.7266218181818182e+16, 
                                 6.610916636363636e+17, 
                                 1.8958227272727274e+17, 
                                 9.626614500000001e+17, 
                                 1.2611768181818182e+16, 
                                 1.377051818181818e+18, 
                                 5.511040909090909e+17, 
                                 1.8492713636363635e+18, 
                                 2.3963863636363636e+16, 
                                 0.005684863636363637, 
                                 0.0002420967818181818, 
                                 5078.07, 
                                 15.016000000000004, 
                                 37.76599999999999, 
                                 0.0010376686363636363],
                        rtol=1e-7)
    
def test_main_results(example_dataset):
    ds = example_dataset
    #load all SO2 results
    so2_default = ds.get_results("so2")
    so2_fit01 = ds.get_results("so2", "f01")
    so2_fit02 = ds.get_results("so2", "f02")
    
    #plot all SO2 results in top left axes object
    
    #now load the other species and plot them into the other axes objects
    bro=ds.get_results("bro")
    o3=ds.get_results("o3")
    oclo=ds.get_results("oclo")
    # Now calculate Bro/SO2 ratios of the time series and plot them with 
    # SO2 shaded on second y axis    
    bro_so2 = bro/so2_default
    oclo_so2 = oclo/so2_default
    o3_so2 = o3/so2_default
    o3_so2_f01 = o3/so2_fit01
        
    vals = [so2_default.mean(),
            so2_default.std(),
            so2_fit01.mean(),
            so2_fit02.mean(),
            o3.mean(),
            o3.std(),
            bro.mean(),
            oclo.mean(), 
            bro_so2.mean(),
            oclo_so2.mean(),
            o3_so2.median(),
            o3_so2_f01.median()]
    
    npt.assert_allclose(actual=vals,
                        desired=[9.626614500000001e+17, 
                                 9.785535879339162e+17, 
                                 1.0835821818181818e+18, 
                                 6.610916636363636e+17, 
                                 1.8492713636363635e+18, 
                                 1.2135385847822042e+18, 
                                 126046170454545.45, 
                                 42836762272727.27, 
                                 0.0001389915245877655, 
                                 7.579933107191676e-05, 
                                 4.046278032004655, 
                                 3.4814157916895985],
                        rtol=1e-7)


# -*- coding: utf-8 -*-
#
# Pydoas is a Python library for the post-analysis of DOAS result data
# Copyright (C) 2017 Jonas Gliß (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the BSD 3-Clause License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
# See BSD 3-Clause License for more details 
# (https://opensource.org/licenses/BSD-3-Clause)

from datetime import datetime, timedelta, date
from typing import Optional, Sequence, Union
from numpy import asarray, full, polyfit, poly1d, logical_and, polyval, ndarray
from pandas import Series, DataFrame
from matplotlib.pyplot import subplots
from matplotlib.dates import DateFormatter

from pydoas.helpers import to_datetime
from pydoas.dataimport import DataImport, ResultImportSetup
            
class DatasetDoasResults(object):
    """A Dataset for DOAS fit results import and processing
    
    Attributes:
        setup (ResultImportSetup): setup specifying all necessary import 
            settings (please see documentation of :class:`ResultImportSetup`
            for setup details)
        raw_results (dict): dictionary containing the imported results

    Args:
        setup (ResultImportSetup): setup specifying all necessary import 
            settings (please see documentation of :class:`ResultImportSetup`
            for setup details)
        init (int): if 1, the raw results will be loaded immediately
        **kwargs: alternative way to setup :attr:`setup` 
            (:class:`ResultImportSetup` object), which is only used in case
            input parameter **setup** is None.

    """
    def __init__(self, setup: ResultImportSetup = None, init: bool = 1, **kwargs):
        
        self.setup = None
        self.raw_results = {}
        
        self.load_input(setup, **kwargs)
        if init:
            self.load_raw_results()
            
    def load_input(self, setup: ResultImportSetup = None, **kwargs):
        """Process input information
        
        Args:
            setup (ResultImportSetup): setup specifying all necessary import 
                settings (please see documentation of :class:`ResultImportSetup`
                for setup details)
            **kwargs: alternative way to setup :attr:`setup` 
                (:class:`ResultImportSetup` object), which is only used in case
                input parameter **setup** is None.   

        """
        if not isinstance(setup, ResultImportSetup):
            setup = ResultImportSetup(**kwargs)
        self.setup = setup
    
    @property
    def base_path(self) -> str:
        """Basepath of resultfiles"""
        return self.setup.base_path
        
    @property
    def start(self) -> datetime:
        """Start date and time of dataset"""
        return self.setup.start
        
    @property
    def stop(self) -> datetime:
        """Stop date and time of dataset"""
        return self.setup.stop
    
    @property
    def dev_id(self) -> str:
        """Device ID of dataset"""
        return self.setup.dev_id
    
    @property
    def import_info(self):
        """Returns information about result import details"""
        return self.setup.import_info
        
    def change_time_ival(self, start: datetime, stop: datetime) -> None:
        """Change the time interval for the considered dataset
        
        Note: 
            Previously loaded results will be deleted

        Args:
            start (datetime): new start time
            stop (datetime): new stop time    
        """
        if isinstance(start, datetime):
            self.start = start
        if isinstance(stop, datetime):
            self.stop = stop
        self.raw_results = {}
        
    def load_raw_results(self) -> bool:
        """Try to load all results as specified in the setup

        This method will try to load all results as specified in the setup. If
        the import setup is not complete, an exception will be raised.
        
        Returns:
            bool: True if data is loaded, False otherwise

        Raises:
            AttributeError: If the import setup is not complete
        """
        if not self.setup.complete():
            raise AttributeError("Import setup is not complete")
        access = DataImport(self.setup)
        if access.data_loaded:
            self.raw_results = access.results
            self.set_start_stop_time()
            return True
        return False
    
    def has_data(self, fit_id: str, species_id: str, start: datetime = None, stop: datetime = None) -> bool:
        """Checks if specific data is available
        
        Args:
            fit_id (str): ID of the fit scenario
            species_id (str): ID of the species
            start (datetime, optional): Start datetime for the data check
            stop (datetime, optional): Stop datetime for the data check
        
        Returns:
            bool: True if data is available, False otherwise
        """
        if not (fit_id in self.raw_results and species_id in self.raw_results[fit_id]):
            return False
        if all([isinstance(x, datetime) for x in [start, stop]]):
            ts = self.raw_results[fit_id]["start"]
            if not any([start < x < stop for x in ts]):
                return False
        return True
    
    def get_spec_times(self, fit: str) -> tuple:
        """Returns start time and stop time arrays for spectra to a given fit
        
        Args:
            fit (str): ID of the fit scenario
        
        Returns:
            tuple: start and stop time arrays for the spectra
        """

        start = asarray(self.raw_results[fit]["start"])
        stop = asarray(self.raw_results[fit]["stop"])
        return start, stop
           
            
    def set_start_stop_time(self):
        """Get start/stop range of dataset"""
        start_acq, stop_acq = [], []
        
        for fit_id in self.raw_results:
            ts, _ = self.get_spec_times(fit_id)
            if len(ts) > 0:
                start_acq.append(ts.min())
                stop_acq.append(ts.max())
        
        self.setup.start = min(start_acq)
        self.setup.stop = max(stop_acq)
        return self.start, self.stop
    
    def get_start_stop_mask(self, fit, start=None, stop=None):
        """Creates boolean mask for data access only in a certain time interval
        """
        if start is None:
            start = self.start
        elif isinstance(start, date): 
            #assume that all spectra from that day are supposed to be imported
            print ("Start time is date, loading results from all spectra of "
                    "that day")
            start = to_datetime(start)
            stop = start + timedelta(days=1)
        if stop is None:
            stop = self.stop
        i, f = self.get_spec_times(fit)
        mask = logical_and(i >= start, i <= stop)
        return mask, i[mask], f[mask]
        
    def get_meta_info(self, fit, meta_id, start=None, stop=None):
        """Get meta info array
        
        :param str meta_id: string ID of meta information
        :param array boolMask: boolean mask for data retrieval
        
        .. note::
        
            Bool mask must have same length as the meta data array
            
        """
        if meta_id not in self.raw_results[fit]:
            return 0
        m, start, stop = self.get_start_stop_mask(fit, start, stop)
        return Series(self.raw_results[fit][meta_id][m], start)
        
    def _get_data(self, species_id, fit_id):
        """Access raw data array
        
        :param str species_id: string ID of species
        :param str fit_id: string ID of fit scenario
    
        """
        return self.raw_results[fit_id][species_id]
        
    def get_results(self, species_id, fit_id=None, start=None, stop=None):
        """Get spectral results object
        
        :param str species_id: string ID of species
        :param str fit_id: string ID of fit scenario (if None, tries to 
            load default fit_id)
        :param start: if valid (i.e. datetime object) only data after that time
            stamp is considered
        :param stop: if valid (i.e. datetime object) only data before that time
            stamp is considered    
        """
        if fit_id is None:
            fit_id = self.setup.default_fit_ids[species_id]
        if not fit_id in self.setup.get_fit_ids():
            return False
        m, start, stop = self.get_start_stop_mask(fit_id, start, stop)
        if not sum(m):
            return False
        dat = self._get_data(species_id, fit_id)[m]
        times = start + (stop - start) / 2
        dat_err = self._get_data((species_id + "_err"), fit_id)[m]
        return DoasResults(dat, times, start, stop, dat_err, species_id, fit_id, 3)

    def get_default_fit_id(self, species_id):
        """Get default fit scenario id for species
        
        :param str species_id: ID of species (e.g. "so2")
        """
        return self.setup.default_fit_ids[species_id]
            
    def set_default_fitscenarios(self, default_dict):
        """Update default fit scenarios for species
        
        :param dict default_dict: dictionary specifying new default fit 
            scenarios, it could e.g. look like::
            
                default_dict = {"so2"   :   "f01",
                                "o3"    :   "f01",
                                "bro"   :   "f03"}
                                
        """
        try:
            self.setup.set_defaults(default_dict)
            return 1
        except:
            return 0
       
    def plot(self,species_id, fit_id=None, start=None, stop=None, **kwargs):
        """Plot DOAS results"""
        res = self.get_results(species_id, fit_id, start, stop)
        if res is not False:
            return res.plot(**kwargs)
        return 0
            
    def scatter_plot(self, species_id_xaxis, fit_id_xaxis, species_id_yaxis,
                     fit_id_yaxis, lin_fit_opt=1, species_id_zaxis=None, 
                     fit_id_zaxis=None, start=None, stop=None, ax=None, 
                     **kwargs):
        """Make a scatter plot of two species
        
        :param str species_id_xaxis: string ID of x axis species (e.g. "so2")
        :param str fit_id_xaxis: fit scenario ID of x axis species (e.g. "f01")
        :param str species_id_yaxis: string ID of y axis species (e.g. "so2")
        :param str fit_id_yaxis: fit scenario ID of y axis species (e.g. "f02")
        :param str species_id_zaxis: string ID of z axis species (e.g. "o3")
        :param str fit_id_zaxis: fit scenario ID of z axis species (e.g. "f01")
        :param datetime start: start time stamp for data retrieval
        :param datetime stop: stop time stamp for data retrieval
        :param ax: matplotlib axes object (None), if provided, then the result 
            is plotted into the axes
        :param kwargs: keyword arguments for matplotlib scatter plot
            (e.g. color, marker, edgecolor, etc.)
        """
        res_x = self.get_results(species_id_xaxis, fit_id_xaxis, start, stop)
        res_y = self.get_results(species_id_yaxis, fit_id_yaxis, start, stop)
        
        res_z = None
        if species_id_zaxis is not None and fit_id_zaxis is not None:
            res_z = self.get_results(\
                        species_id_zaxis, fit_id_zaxis, start, stop)
        if ax is None:
            fig, ax = subplots(1, 1)
        else:
            fig = ax.figure.canvas.figure
        if res_z is not None:
            sc = ax.scatter(res_x.values, res_y.values, 15, res_z.values,\
                                marker = 'o', edgecolor = 'none', **kwargs)
            cb = fig.colorbar(sc, ax = ax, shrink = 0.9, **kwargs)    
            cb.set_label(species_id_zaxis + " " + fit_id_zaxis, **kwargs)
        else:
            ax.plot(res_x.values, res_y.values, " b*", label = "Data",\
                                                                **kwargs)
        ax.set_xlabel(species_id_xaxis + " " + fit_id_xaxis)
        ax.set_ylabel(species_id_yaxis + " " + fit_id_yaxis)
        ax.grid()
        if lin_fit_opt:
            self.linear_regression(res_x.values, res_y.values, ax = ax)
        ax.legend(loc = 'best', fancybox = True, framealpha = 0.5, fontsize = 12)
        return ax 
    
    def linear_regression(self, x_data, y_data, mask = None, ax = None):
        """Perform linear regression and return parameters
        
        :param ndarray x_data: x data array
        :param ndarray y_data: y data array
        :param ndarray mask: mask specifying indices of input data supposed to 
            be considered for regression (None)
        :param ax: matplotlib axes object (None), if provided, then the result 
            is plotted into the axes
        """
        if mask is None:
            mask = full(len(y_data), True, dtype=bool)
        poly = poly1d(polyfit(x_data[mask], y_data[mask], 1))

        if ax is not None:
            ax.plot(x_data, polyval(poly, x_data), "--r",\
                                label = "Slope: %.2f" %(poly[1]))
        return poly
    
    def get_fit_import_setup(self):
        """Get the current fit import setup"""
        return self.setup.import_info

    def __getitem__(self, key):
        """Get attribute using bracketed syntax"""
        for k,v in list(self.__dict__.items()):
            if k == key:
                return v
            try:
                return v[key]
            except:
                pass

    def __setitem__(self, key, val):
        """Change attribute value using bracket syntax"""
        for k,v in list(self.__dict__.items()):
            if k == key:
                self.__dict__[key] = val
                return               
       
    def __str__(self):
        """String representation"""
        return ("\nDOAS result dataset\n-------------------\n" + str(self.setup))

class DoasResults(Series):
    """Data time series for handling and analysing DOAS fit results
    
    :param arraylike data: DOAS fit results (column densities)
    :param arraylike index: Time stamps of data points
    :param arraylike fit_errs: DOAS fit errors
    :param string species_id: String specifying the fitted species
    :param string fit_id: Unique string specifying the fit scenario used
    :param int fit_errs_corr_fac: DOAS fit error correction factor 
    """
    def __init__(
            self, 
            data: Union[Sequence, ndarray, Series], 
            index: Optional[Union[Sequence, ndarray]] = None, 
            start_acq: Optional[Union[Sequence, ndarray]] = None, 
            stop_acq: Optional[Union[Sequence, ndarray]] = None,
            fit_errs: Optional[Union[Sequence, ndarray]] = None, 
            species_id: Optional[str] = None, 
            fit_id: Optional[str] = None,
            fit_errs_corr_fac=1.0):
        if start_acq is None:
            start_acq = []
        if stop_acq is None:
            stop_acq = []
        if species_id is None:
            species_id = ""
        if fit_id is None:
            fit_id = ""
        if isinstance(data, Series):
            index = data.index
            species_id = data.name
            data = data.values
        super().__init__(data, index, name=species_id)
        if fit_errs is None:
            fit_errs = []
            
        self.fit_errs = fit_errs

        self.fit_id = fit_id        
        self.fit_errs_corr_fac = fit_errs_corr_fac
        self.start_acq = start_acq
        self.stop_acq = stop_acq
         
    @property
    def start(self):
        """Start time of data"""
        return self.index[0]
    
    @property    
    def stop(self):
        """Stop time of data"""
        return self.index[-1]
        
    @property
    def species(self):
        """Return name of current species"""
        return self.name
    
    def has_start_stop_acqtamps(self):
        """Checks if start_time and stop_time arrays have valid data"""
        try:
            if not all([isinstance(x, datetime) for x in self.start_acq]):
                raise Exception("Invalid value encountered in start_acq")
            if not all([isinstance(x, datetime) for x in self.stop_acq]):
                raise Exception("Invalid value encountered in stop_acq")
            if not all([len(self) == len(x) for x in [self.start_acq, self.stop_acq]]):
                raise Exception("Lengths of arrays do not match...")
            return True
        except Exception:
            return False
       
    def merge_other(self, other, itp_method="linear", dropna=True):
        """Merge with other time series sampled on different grid
        
        Note
        ----
        
        This object will not be changed, instead, two new Series objects will
        be created and returned
        
        Parameters
        ----------
        other : Series
            Other time series 
        itp_method : str
            String specifying interpolation method (e.g. linear, quadratic)
        dropna : bool
            Drop indices containing NA after merging and interpolation 
            
        Returns
        -------
        tuple
            2-element tuple containing
            
            - this Series (merged)
            - other Series (merged)
        
        """
        if not isinstance(other, Series):
            raise ValueError("Need pandas Series instance (or objects "
                            "inheriting from it)")
        df = DataFrame(dict(s1=self,s2=other)).interpolate(itp_method)
        if dropna:
            df = df.dropna()
        return df.s1, df.s2
        
    def get_data_above_detlim(self):
        """Get fit results exceeding the detection limit 
        
        The detection limit is determined as follows::
        
            self.fit_errs_corr_fac*self.data_err
            
        """
        return self[self > self.fit_errs * self.fit_errs_corr_fac]
        
    def plot(self, date_fmt=None, **kwargs):
        """Plot time series
        
        Uses plotting utility of :class:`Series` object (pandas)        
        
        :param **kwargs: - keyword arguments for pandas plot method
        """
        if not "style" in kwargs:
            kwargs["style"]="--x"
        
        try:
            self.index = self.index.to_pydatetime()
        except:
            pass
        ax = super(DoasResults, self).plot(**kwargs)
        try:
            if date_fmt is not None:
                ax.xaxis.set_major_formatter(DateFormatter(date_fmt))
        except:
            pass
        ax.set_ylabel(self.species)
        return ax
    
    def shift(self, timedelta=timedelta(0.)):
        """Shift time stamps of object
        
        :param timedelta timedelta: temporal shift 
        :returns: shifted :class:`DoasResults` object 
        """
        new = DoasResults(self.values, 
                          self.index + timedelta, 
                          self.start_acq, 
                          self.stop_acq, 
                          self.fit_errs, self.name, self.fit_id, 
                          self.fit_errs_corr_fac)
        if self.has_start_stop_acqtamps:
            new.start_acq += timedelta
            new.stop_acq += timedelta
        return new
    
    def _add__(self, inp):
        s = super(DoasResults, self).__add__(inp)
        return DoasResults(s.values, s.index)   
        
    def __sub__(self, inp):
        s = super(DoasResults, self).__sub__(inp)
        return DoasResults(s.values, s.index)
    
    def __div__(self, inp):
        s = super(DoasResults, self).__div__(inp)
        return DoasResults(s.values, s.index)
    
    def __mul__(self, inp):
        s = super(DoasResults, self).__mul__(inp)
        return DoasResults(s.values, s.index)    
        
    



# -*- coding: utf-8 -*-
#
# Pydoas is a Python library for the post-analysis of DOAS result data
# Copyright (C) 2017 Jonas Gliß (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the BSD 3-Clause License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
# See BSD 3-Clause License for more details 
# (https://opensource.org/licenses/BSD-3-Clause)
from importlib import metadata
__version__ = metadata.version("pydoas")

from os.path import abspath, dirname

_LIBDIR = abspath(dirname(__file__))

from .inout import *
from . import helpers
from .dataimport import ResultImportSetup
from .analysis import DatasetDoasResults, DoasResults




import os
import matplotlib
from importlib.metadata import version as get_version

# Ensure matplotlib works in headless mode
if os.getenv("SPHINX_BUILD"):
    matplotlib.use('agg')

# Project information
project = "pydoas"
author = "Jonas Gliss"
copyright = "2016, Jonas Gliss"

# Versioning
version = get_version("pydoas")
release = version

# Sphinx extensions
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",
    "sphinx.ext.graphviz",
    "sphinx.ext.napoleon",  # Standard Napoleon
    "sphinx_autodoc_typehints",  # Improved type hints
]

# Paths
templates_path = ["_templates"]
exclude_patterns = ["_build"]

# Language
language = "en"

# Autodoc settings
autodoc_default_options = {
    "members": True,
    "undoc-members": True,
    "private-members": False,
    "special-members": "__init__",
    "show-inheritance": True,
}

# HTML output settings
html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]

# Intersphinx mappings
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
}

# Custom function to handle skipping __init__
def skip(app, what, name, obj, skip, options):
    if name in {"__init__", "__call__"}:
        return False
    return skip

def setup(app):
    app.connect("autodoc-skip-member", skip)



# -*- coding: utf-8 -*-
#
# Pydoas is a Python library for the post-analysis of DOAS result data
# Copyright (C) 2017 Jonas Gliß (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the BSD 3-Clause License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
# See BSD 3-Clause License for more details 
# (https://opensource.org/licenses/BSD-3-Clause)
"""
This module contains I/O routines for DOAS result files
"""
from os.path import join
from os import listdir
from collections import OrderedDict as od
from typing import Optional

def get_data_files(which = "doasis"):
    """Get all example result files from package data"""
    from pydoas import _LIBDIR
    if which == "doasis":
        p = join(_LIBDIR, join("data", "doasis_resultfiles"))
    elif which == "fake":
        p = join(_LIBDIR, join("data", "fake_resultfiles"))
    else:
        raise ValueError("No example resultfiles available for ID %s, "
            "choose from *fake* or *doasis*")
    return listdir(p), p

def get_result_type_ids():
    """Read file import_info.txt and find all valid import types"""
    try:
        from pydoas import _LIBDIR
    except:
        raise
    with open(join(_LIBDIR, join("data", "import_info.txt"))) as f:
        types = []
        for line in f: 
            spl = line.split(":", 1)
            if spl[0] == "type":
                tp = spl[1].split("#")[0].strip()
                if len(tp) > 0:
                    types.append(tp)
    return types

def import_type_exists(type_id):
    """Checks if data import type exists in import_info.txt

    :param str type_id: string ID to be searched in import_info.txt
    """
    if type_id in get_result_type_ids():
        return True
    return False
    
def get_import_info(resulttype="doasis"):
    """Try to load DOAS result import specification for default type 
    
    Import specifications for a specified data type (see package data
    file "import_info.txt" for available types, use the instructions in 
    this file to create your own import setup if necessary)
    
    :param str resulttype: name of result type (field "type" in 
        "import_info.txt" file)
    
    """

    from pydoas import _LIBDIR
    from codecs import decode
    dat = od()
    with open(join(_LIBDIR, join("data", "import_info.txt"))) as f:
        found = 0
        for line in f: 
            if "ENDTYPE" in line and found:
                return dat
            spl = line.split(":", 1)
            if spl[0] == "type" and spl[1].split("#")[0].strip() ==\
                                                            resulttype:
                    found = 1
            if found:
                if not any([line[0] == x for x in["#","\n"]]):
                    k = spl[0].strip()
                    d = [x.strip() for x in spl[1].split("#")[0].split(',')]
                    if k == "time_str_formats":
                        dat[k] = d

                    elif k == "delim":
                        dat[k] = decode(d[0], "unicode-escape")
                    else:
                        try:
                            val = int(d[0])
                        except:
                            val = str(d[0])
                        dat[k] = val
                
    raise IOError("Info for type %s could not be found" %resulttype)

def import_info_file():
    """Return path to supplementary file import_info.txt"""
    from pydoas import _LIBDIR
    return join(_LIBDIR, join("data", "import_info.txt"))
    
def write_import_info_to_default_file(import_dict: dict, file: Optional[str] = None):
    try:
        if import_type_exists(import_dict["type"]):
            raise TypeError("Import specifications for ID %s already exists in "
                "file import_info.txt, please change ID and try again"
                %import_dict["type"])
    except KeyError:
        raise KeyError("Please specify type in dictionary")
    keys = list(get_import_info().keys())
    if file is None:
        file = import_info_file()
    with open(file, "a") as myfile:
        myfile.write("\n\nNEWTYPE\n")
        for k, v in list(import_dict.items()):
            if k in keys:
                if isinstance(v, list):
                    s = str(v[0])
                    for i in range(1, len(v)):
                        s += ",%s" %v[i]
                    myfile.write("%s:%s\n" %(k, s))
                else:
                    myfile.write("%s:%s\n" %(k, v))
            else:
                print(("INVALID KEY (not added) %s: %s" %(k,v)))
        myfile.write("ENDTYPE")
        myfile.close()       
    
        




# -*- coding: utf-8 -*-
"""pydoas example script 2

In this script, a new import type is defined aiming to import the fake
data files in the package data folder "./data/fake_resultfiles
In this example, the import is performed based on specifications of the 
data columns rather than based on searching the header lines of the 
resultfiles.

After import, the an exemplary scatter plot of fake species 3 is performed
for the two fit IDs specified.

"""
import pydoas
from os.path import join
from collections import OrderedDict as od

from SETTINGS import SAVE_DIR, SAVEFIGS, ARGPARSER, DPI, FORMAT

def main():
    ### create some fake results:
    
    ### Get data path
    _, path = pydoas.inout.get_data_files(which="fake")
    
    ### Specify default import parameters 
    # (this does not include resultfile columns of DOAS fit results)
    meta_import_info =  od([("type", "fake"),
                            ("access_type", "col_index"),
                            ("has_header_line" , 1),
                            ("time_str_formats", ["%Y%m%d%H%M"]),
                            ("file_type", "csv"),
                            ("delim", ";"),
                            ("start", 0), #col num
                            ("stop", 1), #col num
                            ("bla" , "Blub"), #invalid (for test purpose)
                            ("num_scans", 4)]) #colnum
    
    # You could create a new default data type now by calling
    from tempfile import NamedTemporaryFile
    with NamedTemporaryFile(mode='w+', delete=False) as temp_file:
        pydoas.inout.write_import_info_to_default_file(meta_import_info, file=temp_file.name)
    # which would add these specs to the import_info.txt file and which
    # would allow for fast access using 
    # meta_info_dict = pydoas.inout.get_import_info("fake")
    
    import_dict = {'species1'   : [2, ['fit1']], #column 2, fit result file 1
                   'species2'   : [2, ['fit2']], #column 2, fit result file 2
                   'species3'   : [3, ['fit1', 'fit2']]} #column 3, fit result file 1 and 2
    
    stp = pydoas.dataimport.ResultImportSetup(path, result_import_dict =\
                import_dict, meta_import_info = meta_import_info)
    
    #: Create Dataset object for setup...
    ds = pydoas.analysis.DatasetDoasResults(stp)
    
    ax = ds.scatter_plot("species3", "fit1", "species3", "fit2",\
                    species_id_zaxis = "species1", fit_id_zaxis = "fit1")
    ax.set_title("Ex.2, scatter + regr, fake species3")
    
    if SAVEFIGS:
        ax.figure.savefig(join(SAVE_DIR, "ex2_out1_scatter.%s" %FORMAT),
                          format=FORMAT, dpi=DPI)

    ### IMPORTANT STUFF FINISHED (Below follow tests and display options)
    
    # Import script options
    options = ARGPARSER.parse_args()
    
    # If applicable, do some tests. This is done only if TESTMODE is active: 
    # testmode can be activated globally (see SETTINGS.py) or can also be 
    # activated from the command line when executing the script using the 
    # option --test 1
    if int(options.test):
        import numpy.testing as npt
        from os.path import basename
        
        npt.assert_array_equal([],
                               [])
        
        # check some propoerties of the basemap (displayed in figure)

        npt.assert_allclose(actual=[],
                            desired=[],
                            rtol=1e-7)
        print("No tests implemented in script: %s" %basename(__file__)) 
    try:
        if int(options.show) == 1:
            from matplotlib.pyplot import show
            show()
    except:
        print("Use option --show 1 if you want the plots to be displayed")
    
if __name__ == "__main__":
    main()


# -*- coding: utf-8 -*-
#
# Pydoas is a Python library for the post-analysis of DOAS result data
# Copyright (C) 2017 Jonas Gliß (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the BSD 3-Clause License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
# See BSD 3-Clause License for more details 
# (https://opensource.org/licenses/BSD-3-Clause)
from datetime import datetime
from os.path import join, exists
from os import listdir
import csv
from warnings import warn
from numpy import asarray, shape 

from pydoas.inout import get_import_info
from pydoas.helpers import to_datetime

class ResultImportSetup(object):
    """
    Setup class for spectral result imports from text like files 
    
    Args:
        base_dir: folder containing resultfiles
        start: time stamp of first spectrum
        stop: time stamp of last spectrum
        meta_import_info: Specify the result file format and columns
            for meta information (see als file import_info.txt or example
            script 2). Input can be str or dict. In case a string is 
            provided, it is assumed, that the specs are defined in 
            import_info.txt, i.e. can be imported (as dictionary) 
            from this file (using :func:`get_import_info`, e.g. with arg = 
            ``doasis``). If a dictionary is provided, the information is 
            directly set from the provided dictionary. 
        result_import_dict: specify file and header information for 
            import. Keys define the used abbreveations after import, the 
            values to each key consist of a list with 2 elements: the first 
            specifies the UNIQUE string which is used to identify this 
            species in the header of a given Fit result file, the second 
            entry is a list with arbitrary length containing the fit 
            scenario IDs defining from which fit scenario result files this 
            specific species is to be extracted.
        
            Example::
            
                result_import_dict = {"so2" : ['SO2_Hermans', ['f01','f02']],
                                      "o3"  : ['o3_burrows'], ['f01']]}
            
            Here ``so2`` and "o3" are imported, the data column in the 
            result files is found by the header string ``'SO2_Hermans'`` / 
            ``'o3_burrows'`` and this species is imported from all fit 
            scenario result files with fit Ids ``["f01", "f02"]`` 
            (UNIQUE substrings in FitScenario file names.
            
            Example file name:
            
                ``D130909_S0628_i6_f19_r20_f01so2.dat``
                
            This (exemplary) filename convention is used for the example 
            result files shipped with this package (see folder 
            pydoas/data/doasis_resultfiles) which include fit result files
            from the software `DOASIS.
            
            The delimiter for retrieving info from these file names is "_", 
            the first substring provides info about the date (day), the 
            second about the start time of this time series (HH:MM), 3rd, 
            4th and 5th information about first and last fitted spectrum 
            number and the corresponding number of the reference spectrum 
            used for this time series and the last index about the fit 
            scenario (fitID).
            
            Each resultfile must therefore include a unique ID in the file
            name by which it can be identified.
        
        default_dict: specify default species,
            e.g.::
        
                dict_like = {"so2"     :   "f02",
                             "o3"      :   "f01"}
                        
        doas_fit_err_factors: fit correction factors 
            (i.e. factors by which the DOAS fit error is increased)::
                
                dict_like = {"so2"     :   "f02",
                             "o3"      :   "f01"}
                             
        dev_id: string ID for DOAS device (of minor importance)
    """
    def __init__(self, base_dir=None, start=datetime(1900, 1, 1),
                 stop=datetime(3000, 1, 1), meta_import_info="doasis",
                 result_import_dict={}, default_dict={},
                 doas_fit_err_factors={}, dev_id=""):
    
        self.base_dir = base_dir
        self._start = None
        self._stop = None
        
        self.start = start
        self.stop = stop
        
        self.dev_id = dev_id
        
        self.import_info = result_import_dict
        self.meta_import_info = {}
        self.default_fit_ids = {}
        self.doas_fit_err_factors = {}
        
        self.check_time_stamps()
        
        self.minimum_meta_keys = [
            "start", 
            "delim", 
            "access_type", 
            "has_header_line", 
            "file_type", 
            "time_str_formats"
            ]
        self.set_defaults(default_dict)
        self.set_fitcorr_factors(doas_fit_err_factors)
        
        if isinstance(meta_import_info, str):
            self.meta_import_info.update(get_import_info(meta_import_info))
        elif isinstance(meta_import_info, dict):
            self.meta_import_info.update(meta_import_info)
        
        if not all([x in list(self.meta_import_info.keys()) for x in self.minimum_meta_keys]):
            raise ImportError(f"Please specify at least the following "
                f"parameters: {self.minimum_meta_keys}, available keys are: {list(self.meta_import_info.keys())}")
        if self.access_type == "header_str" and not self.meta_import_info["has_header_line"]:
            raise ValueError("Invalid combination of result file settings: has_header_line == False and access_type == header_str")
            
    @property
    def start(self):
        """Start time-stamp of data"""
        return self._start
        
    @start.setter
    def start(self, val):
        try:
            self._start = to_datetime(val)
        except:
            warn(f"Input {val} could not be assigned to start time in setup")
    
    @property
    def stop(self):
        """Stop time-stamp of data"""
        return self._stop
        
    @stop.setter
    def stop(self, val):
        try:
            self._stop = to_datetime(val)
        except:
            warn(f"Input {val} could not be assigned to stop time in setup")
    
    @property
    def base_path(self):
        """Old name of base_dir for versions <= 1.0.1"""
        return self.base_dir
        
    def set_start_time(self, dt):
        """Set the current start time
        
        :param datetime dt: start time of dataset
        
        """
        if not isinstance(dt, datetime):
            return False
        self.start = dt
        self.check_time_stamps()
        return True
    
    def set_stop_time(self, dt):
        """Set the current start time
        
        :param datetime dt: start time of dataset
        
        """
        if not isinstance(dt, datetime):
            return False
        self.stop = dt
        self.check_time_stamps()
        return True
        
    def check_time_stamps(self):
        """Check if time stamps are valid and if not, set"""
        if not isinstance(self.start, datetime):
            self.start = datetime(1900, 1, 1) #start time of represented data
        if not isinstance(self.stop, datetime):
            self.stop = datetime(3000, 1, 1) #stop time of represented data  
    
    def complete(self):
        """Checks if basic information is available"""
        exceptions = []
        try:
            if not exists(self.base_dir):
                raise AttributeError("Base directory does not exist")
        except Exception as e:
            exceptions.append(repr(e))
        try:
            if not all([isinstance(x, datetime) for x in [self.start, self.stop]]):
                raise AttributeError("Invalid start / stop timestamps")
        except Exception as e:
            exceptions.append(repr(e))
        try:
            if not bool(self.import_info):
                raise AttributeError("No species import information specified")
        except Exception as e:
            exceptions.append(repr(e))
        if len(exceptions) > 0:
            for exc in exceptions:
                warn(exc)
            return False
        return True
        
    def set_defaults(self, dict_like):
        """Update default fit IDs for fitted species
        
        Scheme::
        
            dict_like = {"so2"     :   "f02",
                         "o3"      :   "f01"}
                      
        """
        fit_ids = self.get_fit_ids()
        if not bool(fit_ids):
            return False
        for species, info in list(self.import_info.items()):
            if species in dict_like:
                v = dict_like[species]
            else:
                v = info[1][0]
            self.default_fit_ids[species] = v
        return True
    
    def set_fitcorr_factors(self, dict_like):
        """Set correction factors for uncertainty estimate from DOAS fit errors
        
        :param dict dict_like: dictionary specifying correction factors for 
            DOAS fit errors (which are usually underestimated, see e.g.
            `Gliss et al. 2015 <http://www.atmos-chem-phys.net/15/5659/
            2015/acp-15-5659-2015.html>`_) for individual fit scenarios, e.g.::
            
                dict_like = {"f01"   :   4.0,
                             "f02"   :   2.0}
        
        Default value is 3.0.
        
        """
        facs = {}        
        for fit_id in self.fit_ids:
            if fit_id in dict_like:
                facs[fit_id] = dict_like[fit_id]
            else:
                facs[fit_id] = 3.0
        self.doas_fit_err_factors = facs
        
    def get_xs_names(self):
        """Set and return the string IDs of all fitted species"""
        xs = []
        for key, val in list(self.import_info.items()):
            if val[0] not in xs:
                xs.append(val[0])
        xs.sort()
        return xs

    def get_fit_ids_species(self, species_id):
        """Find all fit scenarios which contain results of species
        
        :param str species_id: string ID of fitted species (e.g. SO2)
        
        """
        if not species_id in list(self.import_info.keys()):
            return 0
        return self.import_info[species_id][1]
    
    @property
    def fit_ids(self):
        """Returns list with all fit ids"""
        return self.get_fit_ids()
    
    @property
    def access_type(self):
        """Return the current setting for data access type"""
        return self.meta_import_info["access_type"]
    
    @property
    def HEADER_ACCESS_OPT(self):
        """Checks if current settings allow column identification from file
        header line"""
        if self.access_type == "header_str" and\
                    self.meta_import_info["has_header_line"]:
            return True
        return False
    
    @property
    def FIRST_DATA_ROW_INDEX(self):
        if self.meta_import_info["has_header_line"]:
            return 1
        return 0
        
    def get_fit_ids(self):
        """Get all fit id abbreveations
        
        Gets all fit ids (i.e. keys of fit import dict ``self.import_info``)        
        """
        ids = []
        for val in list(self.import_info.values()):
            sublist = val[1]
            for substr in sublist:
                if substr not in ids:
                    ids.append(substr)
        ids.sort()
        return ids
    
    def __getitem__(self, key):
        """Get a class attribute using bracketed syntax"""
        if key in self.__dict__:
            return self.__dict__[key]
    
    def __setitem__(self, key, val):
        """Set a class attribute using bracketed syntax"""
        if key in self.__dict__:
            self.__dict__[key] = val

    def __str__(self):
        """String representation of this class"""
        s = (f"\nSetup\n---------\n\n"
             f"Base path: {self.base_dir}\n"
             f"Start: {self.start}\n"
             f"Stop: {self.stop}\n")
        s = s + "\n Absorption cross sections\n"
        for key, val in list(self.import_info.items()):
            s = s + f"{key}: {val[0]}\n"
        s = s + f"Fit scenario IDs\n{self.fit_ids}\n"
            
        return s
           
class DataImport(object):
    """A class providing reading routines of DOAS result files 
    
    Here, it is assumed, that the results are stored in FitResultFiles, 
    tab delimited whereas the columns correspond to the different variables 
    (i.e. fit results, metainfo, ...) and the rows to the individual 
    spectra.
    """
    def __init__(self, setup = None):
        if not isinstance(setup, ResultImportSetup):
            setup = ResultImportSetup()
        self.setup = setup
        
        self.file_type = None
        self.delim = "\t"
        
        self.species_pre_string = ""
        
        self._import_conv_funcs = {"fit_err_add_col" : int}
        
        self._time_str_index = 0
        self.time_str_formats = []
        #:specifying string ids and setup
        self._meta_ids = {el:None for el in list(self._type_dict.keys())}
        
        self.file_paths = {}
        self.results = {}
        
        self.data_loaded = self.get_data()
    
    @property
    def _type_dict(self):
        """Return dictionary specifying data types of import parameters"""
        return {"start"                 :   datetime.strptime,
                "stop"                  :   datetime.strptime,
                "texp"                  :   float,
                "num_scans"             :   float,
                "azim"                  :   float,
                "elev"                  :   float,
                "lat"                   :   float,
                "lon"                   :   float,
                "geom"                  :   str,
                "chi2"                  :   float,
                "delta"                 :   float,
                "rms"                   :   float,
                "fit_low"               :   float,
                "fit_high"              :   float}
    
    def get_data(self):
        """Load all data"""
        if not exists(self.base_dir):
            raise IOError("DOAS data import failed: result base directory "
                "%s does not exist" %self.base_dir)
        self.load_result_type_info()
        self.get_all_files()
        self.load_results()
        return True
    
    def load_result_type_info(self):
        """Load import information for result type specified in setup
        
        The detailed import information is stored in the package data file
        import_info.txt, this file can also be used to create new filetypes
        """
        info = self.setup.meta_import_info
        for k,v in list(info.items()):
            self[k] = v
    
    def _update_attribute(self, key, val):
        """Update one of the attributes
        
        Checks if key is class attribute or - alternatively - a valid meta 
        key (i.e. key of ``self._meta_ids``). If so, checks if a 
        specific data type for this attribute is required (in 
        ``self._type_dict``). If the latter is the case, the input value 
        will only be set if it can be converted given the specific type 
        conversion. 
        
        :param str key: valid key of class or ``self._meta_ids`` dict
        :param val: new value
        
        """
        if key in self._import_conv_funcs:
            try:
                val = self._import_conv_funcs[key](val)
            except:
                return False
            
        if key in self.__dict__:
            self.__dict__[key] = val
            return True
        if key in self._meta_ids:
            self._meta_ids[key] = val
            return True
        return False
        
    @property
    def base_dir(self):
        """Returns current basepath of resultfiles"""
        return self.setup.base_dir
        
    @property
    def start(self):
        """Returns start date and time of dataset"""
        return self.setup.start
        
    @property
    def stop(self):
        """ Returns stop date and time of dataset"""
        return self.setup.stop
    
    @property
    def time_str_format(self):
        """Returns datetime formatting info for string to datetime conversion
        
        This information should be available in the resultfile type 
        specification file (package data: data/import_info.txt)        
        """
        return self.time_str_formats[self._time_str_index]
    
    @property
    def fit_err_add_col(self):
        """Return current value for relative column of fit errors"""
        try:
            return self.setup.meta_import_info["fit_err_add_col"]  
        except:
            return 0
            
    def init_result_dict(self):
        """Initiate the result dictionary"""
        res = {}
        for fit_id in self.setup.fit_ids:
            res[fit_id] = {} 
            for meta_id in self._meta_ids:
                res[fit_id][meta_id]=[]
            for species, info in list(self.setup.import_info.items()):
                if fit_id in info[1]:
                    res[fit_id][species] = []
                    res[fit_id][species + "_err"] = []
        self.results = res
    
    def find_valid_indices_header(self, fileheader, dict):
        """Find positions of species in header of result file
        
        :param list fileheader: header row of resultfile
        :param dict dict: dictionary containing species IDs (keys) and the
            corresponding (sub) strings (vals) to find them in the header
    
        """
        ind = {}
        for key, val in list(dict.items()):
            v = self.find_col_index(val, fileheader)
            if v != -1:
                ind[key] = v
        return ind
    
    def find_all_indices(self, fileheader, fit_id):
        """Find all relevant indices for a given result file (fit scenario)
        
        :param list fileheader: list containing all header strings from 
            result file (not required if data access mode is from columns
            see also :func:`HEADER_ACCESS_OPT` in 
            :class:`ResultImportSetup`)
        :param str fit_id: ID of fit scenario (required in order to find 
            all fitted species supposed to be extracted, specified in 
            ``self.setup.import_info``)
        """
        #load all metainfo indices
        warnings = []
        if self.setup.HEADER_ACCESS_OPT:
            ind = self.find_valid_indices_header(fileheader, self._meta_ids)
        else:
            ind = {}
            for key, val in list(self._meta_ids.items()):
                try:
                    ind[key] = int(val)
                except:
                    pass
        #now find indices of all species supposed to be extracted from this fit 
        #scenario
        for species, info in list(self.setup.import_info.items()):
            if fit_id in info[1]:
                if self.setup.HEADER_ACCESS_OPT:
                    substr = self.species_pre_string + info[0]
                    idx = self.find_col_index(substr, fileheader)
                else:
                    idx = info[0] 
                if idx != -1:                      
                    ind[species] = idx
                    ind[species + "_err"] = ind[species] +\
                                                self.fit_err_add_col
                else:
                    warnings.append("Failed to find column index for "
                        "species %s using header search string %s" 
                        %(species, substr))
        return ind, warnings
        
    def load_results(self):
        """Load all results 
        
        The results are loaded as specified in ``self.import_setup`` for all 
        valid files which were detected in :func:`get_all_files` which 
        writes ``self.file_paths``
        """
        # delete all previous results
        self.init_result_dict()
        all_warnings = []
        # loop over different fit IDs
        for fit_id in self.setup.fit_ids:
            # loop over all files corresponding to one fit ID
            for file in self.file_paths[fit_id]:
                data = self.read_text_file(file)
                last_index = shape(data)[0]
                #load all metainfo indices
                ind, warnings = self.find_all_indices(data[0], fit_id)
                if bool(warnings):
                    all_warnings.append(warnings)

                #Here, the import begins (loop over data rows in file)
                for k in range(self.setup.FIRST_DATA_ROW_INDEX, last_index):
                    start = datetime.strptime(data[k][ind["start"]],\
                                              self.time_str_format)
                    if self.start <= start <= self.stop:
                        for key, index in list(ind.items()):
                            try:
                                if key in ["start", "stop"]:
                                    _t=datetime.strptime(data[k][index], self.time_str_format)
                                    self.results[fit_id][key].append(_t)
                                else:
                                    #try to convert the entry into float
                                    self.results[fit_id][key].append(float(data[k][index]))
                            except:
                                self.results[fit_id][key].append(data[k][index])
                                
                                
        for key, dic in list(self.results.items()):
            for k, lst in list(dic.items()):
                self.results[key][k] = asarray(lst)
        
        if bool(all_warnings):
            for warning_list in all_warnings:
                for warning in warning_list:
                    print(warning)
        
    def find_col_index(self, substr, header):
        """Find the index of the column in data
        
        :param str substr: substr identifying the column in header
        :param list header: the header of the data in which index of substr is searched
        
        """
        try:
            return next((i for i, s in enumerate(header) if substr in s), -1)
        except Exception:
            return -1
    
    def _update_time_str_format(self, data):
        """Set the format strings for datetime info in the result-files"""
        func = self._type_dict["start"]
        if self.setup.HEADER_ACCESS_OPT:
            col = self.find_col_index(self._meta_ids["start"], data[0])
        else:
            col = self.setup.meta_import_info["start"]
        if col == -1:
            return 0
        fmts = self.time_str_formats
        for k in range(len(fmts)):
            try:
                func(data[self.setup.FIRST_DATA_ROW_INDEX][col],fmts[k])
                self._time_str_index = k
                return 1
            except:
                pass
        return 0
        
    def check_time_match(self, data):
        """Check if data is within time interval set by self.start and self.stop
        
        :param list data: data as read by :func:`read_text_file`
        :returns: - bool, Match or no match
        """
        func = self._type_dict["start"]
        if self.setup.HEADER_ACCESS_OPT:
            col = self.find_col_index(self._meta_ids["start"], data[0])
        else:
            col = self.setup.meta_import_info["start"]
        if col == -1:
            return 0
        if self.start is None or self.stop is None:
            return 1
        last_index = shape(data)[0]
        #loop over all fitted spectra in the file and find matches
        for k in range(self.setup.FIRST_DATA_ROW_INDEX, last_index):
            t = func(data[k][col], self.time_str_format)
            if self.start < t < self.stop:
                return 1
        return 0
    
    @property
    def first_file(self):
        """Get filepath of first file match in ``self.base_dir``
        
        This can for instance be read with :func:`read_text_file`
        """
        all = [join(self.base_dir, f) for f in listdir(self.base_dir) if\
                                                    f.endswith(self.file_type)]
        return all[0]
    
    def init_filepaths(self):
        """Initate the file paths"""
        for fit_id in self.setup.fit_ids:
            self.file_paths[fit_id] = []
            
    def get_all_files(self):
        """Get all valid files based on current settings
        
        Checks ``self.base_dir`` for files matching the specified file type, 
        and which include one of the required fit IDs in their name. Files
        matching these 2 criteria are opened and the spectrum times are read 
        and checked. If they match the time interval specified by
        ``self.start`` and ``self.stop`` the files are added to the dictionary 
        ``self.file_paths`` where the keys specify the individual fit scenario 
        IDs.
        
        .. note::
        
            This function does not load data but only assigns the individual
            result files to the fit IDs, the data will then be loaded calling
            :func:`load_results`
        
        """
        if not exists(self.base_dir):
            raise IOError("DOAS data import failed: result base directory "
                "%s does not exist" %self.base_dir)
        self.init_filepaths()
        all_files = [join(self.base_dir, f) for f in listdir(\
                         self.base_dir) if f.endswith(self.file_type)]
        if not len(all_files) > 0:
            raise IOError("Data import failed, no files of type %s could "
                "be found in current base folder: %s" #
                %(self.file_type, self.base_dir))
        #first check if time conversion works
        self._update_time_str_format(self.read_text_file(all_files[0]))
        for fname in all_files: 
            for fit_id in self.setup.fit_ids:
                if fname.find(fit_id) >- 1:
                    data = self.read_text_file(fname)
                    found = self.check_time_match(data)
                    if found:
                        self.file_paths[fit_id].append(fname)
        
        del data, all_files
     
    def read_text_file(self, p):
        """Read text file using csv.reader and return data as list
        
        :param str p: file path
        :returns list: data 
        """
        with open(p) as f:
            reader = csv.reader(f, delimiter=str(self.delim))
            data = list(reader)
        return data
    
    def __setitem__(self,key, val):
        """Set item method"""
        self._update_attribute(key, val)





# -*- coding: utf-8 -*-
"""pydoas example script 1

Introductory script illustrating how to import data from DOASIS 
resultfiles. The DOASIS result file format is specified as default in the
package data file "import_info.txt". 
Creates a result data set using DOASIS example data and plots some examples

"""
import pydoas
import matplotlib.pyplot as plt
from os.path import join

from SETTINGS import SAVE_DIR, SAVEFIGS, ARGPARSER, DPI, FORMAT


def main():
    plt.close("all")
    ### Get example data base path and all files in there
    _, path = pydoas.get_data_files("doasis")
    
    ### Device ID of the spectrometer (of secondary importance)
    dev_id = "avantes"
    
    ### Data import type (DOASIS result file format)
    res_type = "doasis"
    
    ### Specify the the import details
    
    # here, 3 x SO2 from the first 3 fit scenario result files (f01, f02, f03) 
    # BrO from f04, 2 x O3 (f02, f04) and OClO (f04)
    import_dict = {
        'so2'  : ['SO2_Hermans_298_air_conv', ['f01','f02','f03']],
        'bro'  : ['BrO_Wil298_Air_conv', ['f04']],
        'o3'   : ['o3_221K_air_burrows_1999_conv', ['f02', 'f04']],
        'oclo' : ['OClO_293K_Bogumil_2003_conv',['f04']]}
    
    ### Specify the default fit scenarios for each species
    
    # After import, the default fit scenarios for each species are used
    # whenever fit scenarios are not explicitely specified
    default_dict = {"so2"  :   "f03",
                    "bro"  :   "f04",
                    "o3"   :   "f04",
                    "oclo" :   "f04"}
    
    #: Create import setup object
    stp = pydoas.dataimport.ResultImportSetup(path, result_import_dict =\
            import_dict, default_dict = default_dict, meta_import_info = res_type,\
            dev_id = dev_id)

    #: Create Dataset object for setup...
    ds = pydoas.analysis.DatasetDoasResults(stp)
    
    #: ... and load results
    ds.load_raw_results()
    
    ### plot_some_examples
    fig1, axes = plt.subplots(2, 2, figsize = (16, 8), sharex = True)
    ax = axes[0,0]
    
    #load all SO2 results
    so2_default = ds.get_results("so2")
    so2_fit01 = ds.get_results("so2", "f01")
    so2_fit02 = ds.get_results("so2", "f02")
    
    #plot all SO2 results in top left axes object
    so2_default.plot(style="-b", ax=ax, label="so2 (default, f03)")
    so2_fit01.plot(style="--c", ax=ax, label="so2 (f01)")
    so2_fit02.plot(style="--r", ax=ax, label="so2 (f02)").set_ylabel("SO2 [cm-2]")
    ax.legend(loc='best', fancybox=True, framealpha=0.5, fontsize=9)
    ax.set_title("SO2")
    fig1.tight_layout(pad = 1, w_pad = 3.5, h_pad = 3.5)
    
    #now load the other species and plot them into the other axes objects
    bro=ds.get_results("bro")
    bro.plot(ax=axes[0, 1], label="bro", title="BrO").set_ylabel("BrO [cm-2]")
    
    o3=ds.get_results("o3")
    o3.plot(ax=axes[1, 0], label="o3",
            title="O3").set_ylabel("O3 [cm-2]")
    oclo=ds.get_results("oclo")
    oclo.plot(ax=axes[1, 1], label="oclo", 
              title="OClO").set_ylabel("OClO [cm-2]")
    
    # Now calculate Bro/SO2 ratios of the time series and plot them with 
    # SO2 shaded on second y axis    
    bro_so2 = bro/so2_default
    oclo_so2 = oclo/so2_default
    
    fig2, axis = plt.subplots(1,1, figsize=(12,8))
    bro_so2.plot(ax=axis, style=" o", label="BrO/SO2")
    oclo_so2.plot(ax=axis, style=" x", label="OClO/SO2")
    #axis.set_ylabel("BrO/SO2")
    so2_default.plot(ax=axis, kind="area",
                     secondary_y=True, alpha=0.3).set_ylabel("SO2 CD [cm-2]")
    axis.legend()
    if SAVEFIGS:
        fig1.savefig(join(SAVE_DIR, "ex1_out1.%s" %FORMAT), 
                     format=FORMAT, dpi=DPI)
        fig2.savefig(join(SAVE_DIR, "ex1_out2.%s" %FORMAT),
                     format=FORMAT, dpi=DPI)
    
    ### IMPORTANT STUFF FINISHED (Below follow tests and display options)
    
    # Import script options
    options = ARGPARSER.parse_args()
    
    # If applicable, do some tests. This is done only if TESTMODE is active: 
    # testmode can be activated globally (see SETTINGS.py) or can also be 
    # activated from the command line when executing the script using the 
    # option --test 1
    if int(options.test):
        ### under development
        import numpy.testing as npt
        import numpy as np
        from os.path import basename
        
        npt.assert_array_equal([len(so2_default),
                                ds.get_default_fit_id("so2"),
                                ds.get_default_fit_id("bro"),
                                ds.get_default_fit_id("oclo")],
                               [22, "f03", "f04", "f04"])
        
        vals = [so2_default.mean(),
                so2_default.std(),
                so2_fit01.mean(),
                so2_fit02.mean(),
                bro.mean(),
                oclo.mean(), 
                bro_so2.mean(),
                oclo_so2.mean(),
                np.sum(ds.raw_results["f01"]["delta"])]
        
        npt.assert_allclose(actual=vals,
                            desired=[9.626614500000001e+17,
                                     9.785535879339162e+17,
                                     1.0835821818181818e+18,
                                     6.610916636363636e+17,
                                     126046170454545.45,
                                     42836762272727.27,
                                     0.0001389915245877655,
                                     7.579933107191676e-05,
                                     0.125067],
                            rtol=1e-7)
        print("All tests passed in script: %s" %basename(__file__)) 
    try:
        if int(options.show) == 1:
            plt.show()
    except:
        print("Use option --show 1 if you want the plots to be displayed")

if __name__=="__main__":
    main()


from traceback import format_exc
import importlib.util
import pathlib
import sys


def get_all_script_paths(script_dir: pathlib.Path, pattern: str, exclude: list[str]) -> list[pathlib.Path]:
    """
    Get all script paths in a directory matching a pattern, excluding specified files.
    
    Parameters
    ----------
    script_dir : pathlib.Path
        The directory to search for scripts.
    pattern : str
        The glob pattern to match script files.
    exclude : list of str
        List of filenames to exclude from the results.
    
    Returns
    -------
    list of pathlib.Path
        List of paths to the scripts that match the pattern and are not excluded.
    """
    
    return [x for x in script_dir.glob(pattern) if not x.name in exclude]

def run_all_scripts(all_intro_scripts: list[pathlib.Path]) -> tuple[list[str], list[str], list[str]]:
    """
    Executes a list of introductory scripts and categorizes their outcomes.

    Parameters
    ----------
    all_intro_scripts : list of pathlib.Path
        A list of paths to the introductory scripts to be executed.

    Returns
    -------
    test_err_messages : list of str
        Messages indicating which scripts failed due to assertion errors and their tracebacks.
    passed_messages : list of str
        Messages indicating which scripts passed all tests.
    crashed_messages : list of str
        Messages indicating which scripts encountered unexpected errors.
    """

    test_err_messages: list[str] = []
    passed_messages: list[str] = []
    crashed_messages: list[str] = []
    for script_path in all_intro_scripts:
        try:
            print(f"Running {script_path.name}")
            run_script(script_path=script_path)
            passed_messages.append(f"All tests passed in script: {script_path.name}")
        except AssertionError as e:
            msg = (f"\n\n"
                f"--------------------------------------------------------\n"
                f"Tests in script {script_path.name} failed.\n"
                f"Error traceback:\n {format_exc(e)}\n"
                f"--------------------------------------------------------"
                f"\n\n"
                )
            test_err_messages.append(msg)
        except Exception as e:
            crashed_messages.append(f"Unexpected error in {script_path.name}: {e}")
    return test_err_messages,passed_messages,crashed_messages

def print_output_runall(options, test_err_messages, passed_messages, crashed_messages):
    """
    Print the results of running all tests and scripts.

    This function prints the results of tests and script executions, including
    test failures, test successes, and crashed messages. The test results are
    printed only if the test mode is active.
    
    Parameters
    ----------
    options : object
        An object containing various options, including the test mode flag.
    test_err_messages : list of str
        A list of error messages from failed tests.
    passed_messages : list of str
        A list of messages from passed tests.
    crashed_messages : list of str
        A list of messages from crashed scripts.
    
    Returns
    -------
    None
    """

    # If applicable, do some tests. This is done only if TESTMODE is active:
    # testmode can be activated globally (see SETTINGS.py) or can also be
    # activated from the command line when executing the script using the
    # option --test 1
    if int(options.test):
        print("\n----------------------------\n"
            "T E S T  F A I L U R E S"
            "\n----------------------------\n")
        if test_err_messages:
            for msg in test_err_messages:
                print(msg)
        else:
            print("None")
        print("\n----------------------------\n"
            "T E S T  S U C C E S S"
            "\n----------------------------\n")
        if passed_messages:
            for msg in passed_messages:
                print(msg)
        else:
            print("None")

    print("\n----------------------------\n"
        "C R A S H E D"
        "\n----------------------------\n")
    if crashed_messages:
        for msg in crashed_messages:
            print(msg)
    else:
        print("None")

def run_script(script_path: pathlib.Path):
    module_name = script_path.stem  # Use filename as module name

    spec = importlib.util.spec_from_file_location(module_name, script_path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module  # Add module to sys.modules
    spec.loader.exec_module(module)  # Execute module
    if hasattr(module, "main"):
        module.main()
    else:
        raise ValueError(f"Please move content of __main__ to main() in {script_path.name}")
    
IGNORE_SCRIPTS = []
SCRIPT_PATTERN = "ex[0-9]*.py"

if __name__ == "__main__":
    from SETTINGS import ARGPARSER, SCRIPTS_DIR
    from time import time

    options = ARGPARSER.parse_args()

    t0 = time()
    all_example_scripts = get_all_script_paths(SCRIPTS_DIR, SCRIPT_PATTERN, IGNORE_SCRIPTS)
    test_err_messages, passed_messages, crashed_messages = run_all_scripts(all_example_scripts)
    t1 = time()    
    print_output_runall(options, test_err_messages, passed_messages, crashed_messages)
    print(f"Total runtime: {t1 - t0:.2f} s")


# -*- coding: utf-8 -*-
#
# Pyplis is a Python library for the analysis of UV SO2 camera data
# Copyright (C) 2017 Jonas Gliß (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License a
# published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
import argparse
import pathlib
from matplotlib import rcParams


rcParams.update({'figure.autolayout': True})
rcParams.update({'font.size': 13})

# if True, some of the actual results of the scripts are verified
TESTMODE = True

SAVEFIGS = 1 # save plots from this script in SAVE_DIR
DPI = 150 #pixel resolution for saving
FORMAT = "png" #format for saving

SCREENPRINT = 0 #show images on screen when executing script

# Directory where results are stored
# Directory where results are stored
SCRIPTS_DIR = pathlib.Path(__file__).parent
SAVE_DIR = SCRIPTS_DIR / "scripts_out"

ARGPARSER = argparse.ArgumentParser()
ARGPARSER.add_argument('--show', default=SCREENPRINT)
ARGPARSER.add_argument('--test', default=TESTMODE)


# -*- coding: utf-8 -*-
#
# Pydoas is a Python library for the post-analysis of DOAS result data
# Copyright (C) 2017 Jonas Gliß (jonasgliss@gmail.com)
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the BSD 3-Clause License
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
# See BSD 3-Clause License for more details 
# (https://opensource.org/licenses/BSD-3-Clause)
"""
Module containing all sorts of helper methods
"""

from datetime import datetime, time, date

def to_datetime(value):
    """Method to evaluate time and / or date input and convert to datetime"""
    if isinstance(value, datetime):
        return value
    elif isinstance(value, date):
        return datetime.combine(value, time())
    elif isinstance(value, time):
        return datetime.combine(date(1900,1,1), value)
    else:
        raise ValueError("Conversion into datetime object failed for input: "
            "%s (type: %s)" %(value, type(value)))       
    
